/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "game.h"

#include "info.h"

game::game() {
	working = true;
	resource = false;
	vector = false;
	sound = true;
	color = true;
	sizex = 0;
	sizey = 0;
	time = 0;
	size = 0;
	type = TYPE_UNDEFINED;
	group = GROUP_UNDEFINED;
	parent = 0;
}

game::game(const string& Aname) : name(Aname) {
	working = true;
	resource = false;
	vector = false;
	sound = true;
	color = true;
	sizex = 0;
	sizey = 0;
	time = 0;
	size = 0;
	type = TYPE_UNDEFINED;
	group = GROUP_UNDEFINED;
	parent = 0;
}

game::game(const game& A) :
	name(A.name), romof(A.romof),cloneof(A.cloneof),
	sampleof(A.sampleof), description(A.description), year(A.year),
	manufacturer(A.manufacturer),
	working(A.working), resource(A.resource),
	vector(A.vector),
	sound(A.sound), color(A.color),
	sizex(A.sizex), sizey(A.sizey),
	group(A.group), type(A.type), time(A.time), size(A.size),
	rzs(A.rzs), szs(A.szs),
 	clone_bag(A.clone_bag), parent(A.parent) {
}

game::~game() {
}

void game::rom_zip_set_add(const string& Afile) const {
	rzs.insert( rzs.end(), string( Afile ) );
}

void game::sample_zip_set_add(const string& Afile) const {
	szs.insert( szs.end(), string( Afile ) );
}

bool game_set::load(FILE* f) {
	info_t token = info_token_get(f);
	while (token!=info_eof) {
		if (token != info_symbol) return false;
		bool isresource = strcmp(info_text_get(),"resource")==0;
		bool isgame = strcmp(info_text_get(),"game")==0;
		if (isgame || isresource) {
			if (info_token_get(f) != info_open) return false;
			game g;
			g.resource_set(isresource);
			token = info_token_get(f);
			while (token != info_close) {
				if (token != info_symbol)
					return false;
				if (strcmp(info_text_get(),"name")==0) {
					if (info_token_get(f) != info_symbol) return false;
					g.name_set( info_text_get() );
				} else if (strcmp(info_text_get(),"description")==0) {
					if (info_token_get(f) != info_string) return false;
					g.description_set( info_text_get() );
				} else if (strcmp(info_text_get(),"manufacturer")==0) {
					if (info_token_get(f) != info_string) return false;
					g.manufacturer_set( info_text_get() );
				} else if (strcmp(info_text_get(),"year")==0) {
					if (info_token_get(f) != info_symbol) return false;
					g.year_set( info_text_get() );
				} else if (strcmp(info_text_get(),"cloneof")==0) {
					if (info_token_get(f) != info_symbol) return false;
					g.cloneof_set( info_text_get() );
				} else if (strcmp(info_text_get(),"romof")==0) {
					if (info_token_get(f) != info_symbol) return false;
					g.romof_set( info_text_get() );
				} else if (strcmp(info_text_get(),"sampleof")==0) {
					if (info_token_get(f) != info_symbol) return false;
					g.sampleof_set( info_text_get() );
				} else if (strcmp(info_text_get(),"driver")==0) {
					if (info_token_get(f) != info_open)  return false;
					token = info_token_get(f);
					while (token != info_close) {
						if (token != info_symbol) return false;
						if (strcmp(info_text_get(),"status")==0) {
							if (info_token_get(f) != info_symbol) return false;
							g.working_set( strcmp( info_text_get(), "preliminary") != 0 );
						} else if (strcmp(info_text_get(),"color")==0) {
							if (info_token_get(f) != info_symbol) return false;
							g.color_set( strcmp( info_text_get(), "preliminary") != 0 );
						} else if (strcmp(info_text_get(),"sound")==0) {
							if (info_token_get(f) != info_symbol) return false;
							g.sound_set( strcmp( info_text_get(), "preliminary") != 0 );
						} else {
							if (info_skip_value(f) == info_error) return false;
						}
						token = info_token_get(f);
					}
				} else if (strcmp(info_text_get(),"video")==0)  {
					if (info_token_get(f) != info_open)  return false;
					token = info_token_get(f);
					while (token != info_close) {
						if (token != info_symbol) return false;
						if (strcmp(info_text_get(),"screen")==0) {
							if (info_token_get(f) != info_symbol) return false;
							g.vector_set( strcmp( info_text_get(),"vector") == 0 );
						} else if (strcmp(info_text_get(),"x")==0) {
							if (info_token_get(f) != info_symbol) return false;
							g.sizex_set( atoi(info_text_get()) );
						} else if (strcmp(info_text_get(),"y")==0) {
							if (info_token_get(f) != info_symbol) return false;
							g.sizey_set( atoi(info_text_get()) );
						} else {
							if (info_skip_value(f) == info_error) return false;
						}
						token = info_token_get(f);
					}
				} else if (strcmp(info_text_get(),"rom")==0) {
					if (info_token_get(f) != info_open)  return false;
					token = info_token_get(f);
					while (token != info_close) {
						if (token != info_symbol) return false;
						if (strcmp(info_text_get(),"size")==0) {
							if (info_token_get(f) != info_symbol) return false;
							g.size_set( g.size_get() + atoi( info_text_get() ) );
						} else {
							if (info_skip_value(f) == info_error) return false;
						}
						token = info_token_get(f);
					}
				} else {
					if (info_skip_value(f) == info_error) return false;
				}
				token = info_token_get(f);
			}
			insert( g );
		} else {
			if (info_skip_value(f) == info_error)
				return false;
		}
		token = info_token_get(f);
	}

	// compute the clone list for every game
	for(iterator i=begin();i!=end();++i) {
		if (i->cloneof_get().length() != 0) {
			iterator j = find( i->cloneof_get() );
			i->parent_set( &*j );
		}
		if (i->parent_get() != 0) {
			const game* j = i->parent_get();
			while (j) {
				j->clone_bag_get().insert(j->clone_bag_get().end(),&*i);
				j = j->parent_get();
			}
		}
	}

	return true;
}

bool game_set::test() const {
	for(game_set::const_iterator i=begin();i!=end();++i) {
		if (i->cloneof_get().length()) {
			game_set::const_iterator j = find( i->cloneof_get() );
			if (j==end()) {
				cerr << "error: missing definition of game " << i->cloneof_get() << endl;
				return false;
			}
		}
		if (i->romof_get().length()) {
			game_set::const_iterator j = find( i->romof_get() );
			if (j==end()) {
				cerr << "error: missing definition of game " << i->romof_get() << endl;
				return false;
			}
		}
	}

	return true;
}

bool game_set::is_tree_rom_of_present(const string& name, merge_t type) const {
	switch (type) {
		case merge_differential : {
			const_iterator i = find( game( name ) );
			while (i!=end()) {
				if (!i->is_present())
					return false;
				if (i->romof_get().length()==0)
					return true;
				i = find( game( i->romof_get() ) );
			}
			return false;
		}
		case merge_any : {
			const_iterator i = find( game( name ) );
			while (i!=end()) {
				if (i->is_present())
					return true;
				if (i->romof_get().length()==0)
					return false;
				i = find( game( i->romof_get() ) );
			}
			return false;
		}
		case merge_no : {
			const_iterator i = find( game( name ) );
			if (i!=end())
				return i->is_present();
			return false;
		}
		case merge_parent : {
			const_iterator i = root_clone_of_get(name);
			if (i!=end())
				return i->is_present();
			return false;
		}
		case merge_disable :
			return true;
	}
	return false;
}

bool game_set::is_game_neogeo(const string& name) const {
	game_set::const_iterator i = find( game( name ) );
	while (i!=end()) {
		if (i->name_get() == "neogeo")
			return true;
		i = find( game( i->romof_get() ) );
	}
	return false;
}

game_set::const_iterator game_set::root_rom_of_get(const string& name) const {
	game_set::const_iterator i = find( game( name ) );
	while (i != end()) {
		if (i->romof_get().length() == 0)
			return i;
		i = find( i->romof_get() );
	}
	return i;
}

game_set::const_iterator game_set::root_clone_of_get(const string& name) const {
	game_set::const_iterator i = find( game( name ) );
	while (i != end()) {
		if (i->cloneof_get().length() == 0)
			return i;
		i = find( i->cloneof_get() );
	}
	return i;
}

bool game_set::is_game_clone_of(const string& name_son, const string& name_parent) const {
	game_set::const_iterator i = find( game( name_son ) );
	while (i!=end()) {
		if (i->name_get() == name_parent)
			return true;
		if (i->cloneof_get().length() == 0)
			return false;
		i = find( game( i->cloneof_get() ) );
	}
	return false;
}

bool game_set::is_game_rom_of(const string& name_son, const string& name_parent) const {
	game_set::const_iterator i = find( game( name_son ) );
	while (i!=end()) {
		if (i->name_get() == name_parent)
			return true;
		if (i->romof_get().length() == 0)
			return false;
		i = find( game( i->romof_get() ) );
	}
	return false;
}

const game* game::clone_best_get() const {
	const game* r = this;
	for(pgame_container::const_iterator i = clone_bag_get().begin();i!=clone_bag_get().end();++i) {
		const game* rr = (*i)->clone_best_get();
		if (game_by_play_less()(*r,*rr)) {
			r = rr;
		}
	}
	return r;
}

string game::group_derived_get() const {
	if (group_get() != GROUP_UNDEFINED)
		return group_get();
	if (parent_get())
		return parent_get()->group_derived_get();
	return GROUP_UNDEFINED;
}

string game::type_derived_get() const {
	if (type_get() != TYPE_UNDEFINED)
		return type_get();
	if (parent_get())
		return parent_get()->type_derived_get();
	return TYPE_UNDEFINED;
}
