/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __SCRTWEAK_H
#define __SCRTWEAK_H

/* Valid dotclock for vga hardware */
#define TWEAK_DOTCLOCK_HIGH 28322000
#define TWEAK_DOTCLOCK_LOW 25175000

/* VGA Port address */
#define TWEAK_ATTRCON_ADDR 0x3c0 /* Attribute Controller */
#define TWEAK_MISC_ADDR 0x3c2 /* Miscellaneous Register */
#define TWEAK_VGAENABLE_ADDR 0x3c3 /*   VGA Enable Register */
#define TWEAK_SEQ_ADDR 0x3c4 /* Sequencer */
#define TWEAK_GRACON_ADDR 0x3ce /* Graphics Controller */
#define TWEAK_CRTC_ADDR 0x3d4 /* Cathode Ray Tube Controller */
#define TWEAK_STATUS_ADDR 0x3da /* Status Register */
#define TWEAK_CHIPSTECH_ADDR 0x3d6 /* Specific SVGA port */

typedef struct {
	unsigned port;
	unsigned char index;
	unsigned char value;
} tweak_register;

struct tweak_crtc {
	unsigned vt; /* Vertical total scanlines */
	unsigned vrs; /* Vertical retrace start scanline */
	unsigned vre; /* Vertical retrace end scanline */
	unsigned vbs; /* Vertical blank start scanline */
	unsigned vbe7; /* Vertical blank end scanline (7 bit regs version) */
	unsigned vbe; /* Vertical blank end scanline */
	unsigned vde; /* Vertical display end scanline */

	unsigned ht; /* Horizontal total clocks */
	unsigned hrs; /* Horizontal retrace start clocks */
	unsigned hre; /* Horizontal retrace end clocks */
	unsigned hbs; /* Horizontal blank start clocks */
	unsigned hbe; /* Horizontal blank end clocks */
	unsigned hde; /* Horizontal display end clocks */

	unsigned size_x; /* X resolution */
	unsigned size_y; /* Y resolution */
	unsigned virtual_x; /* X virtual resolution */
	unsigned virtual_y; /* Y virtual resolution */
	unsigned bytes_per_scanline; /* bytes per scanline */
	unsigned bytes_per_clock; /* byte per clock */
	unsigned partial_y; /* Unused but shown rows */

	int hsync; /* Horizontal sync polarity 1==negative, 0==positive */
	int vsync; /* Vertical sync polarity 1==negative, 0==positive */

	int is_doublescan; /* boolean, hardware bit */
	unsigned memory_mode; /* 1==byte mode, 2==word mode, 4 dword mode */

	int is_linear; /* boolean, ==0 access by 4 plane (unchained), !=0 linear access (chained) */

	unsigned clocks_per_scanline; /* Number of clocks per scanline */
	unsigned pixels_per_clock; /* Number of pixels per clock */
	unsigned dots_per_clock; /* Number of dots per clock */
	unsigned scanlines_per_line; /* Number of scanlines for one chars row, include char_size_y and doublescan */
	unsigned bits_per_pixel; /* Number of bits per pixel */

	int is_graphics_mode; /* ==0 text, ==1 graphics mode */

	unsigned char_size_x; /* size of char in dot */
	unsigned char_size_y; /* size of char in scanline */

	int dotclock_middle; /* Dot clock is Master clock/2 */
	int dotclock_master; /* Master clock 0==25 1==28 */

	float pixel_clock; /* Pixel frequency in Hz */
	float dot_clock; /* Dot frequency in Hz */
	float vert_clock; /* Vertical frequency in Hz */
	float horz_clock; /* Horizontal frequency in Hz */
};

void tweak_ready_vga_regs(void);
void tweak_display_enable(void);

/* Functions used for computing register values when a division is required */
static __inline__ unsigned tweak_div_low(unsigned value, unsigned value_div) {
	return value / value_div;
}

static __inline__ unsigned tweak_div_high(unsigned value, unsigned value_div) {
	return (value + value_div - 1) / value_div;
}

void tweak_reg_out(unsigned value, unsigned port, int port_index);
unsigned tweak_reg_in(unsigned port, int port_index);

int tweak_reg_write(unsigned value, unsigned port, int port_index, void* arg);
unsigned tweak_reg_read(unsigned port, int port_index, void* arg);

void tweak_reg_array_out(const tweak_register *r, int n);
void tweak_crtc_get(struct tweak_crtc* info, unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);

int tweak_hde_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
unsigned tweak_hde_get(unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_hbs_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
unsigned tweak_hbs_get(unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_hrs_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_hre_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_hbe_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
unsigned tweak_hbe_get(unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_htt_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);

int tweak_vtt_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_vde_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_vbs_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_vbe_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_vrs_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_vre_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
unsigned tweak_vde_get(unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);

int tweak_doublescan_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_dotclock_middle_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_masterclock_input_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_char_size_x_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_char_size_y_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);

int tweak_hsync_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);
int tweak_vsync_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);

int tweak_offset_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);

int tweak_chained_set(int chained, int (*setreg)(unsigned value, unsigned port, int port_index, void* arg), unsigned (*getreg)(unsigned port, int port_index, void* arg), void* arg);

#endif
